/* 
   Unix SMB/CIFS mplementation.
   NDS LDAP help functions for SAMBA
   Copyright (C) Vince Brimhall			2004
    
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   
*/

#include "includes.h"

#include <lber.h>
#include <ldap.h>

#include "smbldap.h"

/* Novell NDS NMAS LDAP extension stuff */
#define NMASLDAP_EXT_LIB						"libnmasext.so"
#define NMASLDAP_GET_PASSWORD					"nmasldap_get_password"
#define NMASLDAP_SET_PASSWORD					"nmasldap_set_password"
#define NMASLDAP_GET_SIMPLE_PWD					"nmasldap_get_simple_pwd"
#define NMASLDAP_PUT_SIMPLE_PWD					"nmasldap_put_simple_pwd"
#define NMASLDAP_GET_PASSWORD_POLICY_DN			"nmasldap_get_password_policy_dn"
#define NMASLDAP_POLICY_CHECK_PASSWORD			"nmasldap_policy_check_password"

void *nmasldap_ext_lib_handle = NULL;

int (*nmasldap_get_password)(LDAP *ld, char *object_dn, int *pwd_len, char *pwd) = NULL;
int (*nmasldap_set_password)(LDAP *ld, char *object_dn, const char *pwd) = NULL;
int (*nmasldap_get_simple_pwd)(LDAP *ld, char *object_dn, int pwd_len, char *pwd) = NULL;
int (*nmasldap_put_simple_pwd)(LDAP *ld, char *object_dn, const char *pwd) = NULL;
int (*nmasldap_get_password_policy_dn)(LDAP *ld, char *object_dn, unsigned int *dnSize, char *dn) = NULL;
int (*nmasldap_policy_check_password)(LDAP *ld, char *object_dn, char *pwd) = NULL;

/**********************************************************************
 Get the user's password from NDS.
 *********************************************************************/
int pdb_nds_get_password(LDAP *ld, char *object_dn, int *pwd_len, char *pwd)
{
	int rc = -1;

	if (nmasldap_get_password != NULL) {
		rc = nmasldap_get_password(ld, object_dn, pwd_len, pwd);
		if (rc == LDAP_SUCCESS) {
#ifdef DEBUG_PASSWORD
			DEBUG(100,("nmasldap_get_password returned %s\n", pwd));
#endif    
			DEBUG(5, ("NDS Universal Password retrieved for %s\n", object_dn));
		} else {
			DEBUG(3, ("nmasldap_get_password for %s failed!\n", object_dn));
		}
	}

	if (rc != LDAP_SUCCESS) {
		if (nmasldap_get_simple_pwd != NULL) {
			rc = nmasldap_get_simple_pwd(ld, object_dn, *pwd_len, pwd);
			if (rc == LDAP_SUCCESS) {
#ifdef DEBUG_PASSWORD
				DEBUG(100,("nmasldap_get_simple_pwd returned %s\n", pwd));
#endif    
				DEBUG(5, ("NDS Simple Password retrieved for %s\n", object_dn));
			} else {
				DEBUG(3, ("nmasldap_get_simple_pwd for %s failed!\n", object_dn));
			}
		}
	}

	if (rc == LDAP_SUCCESS) {
		char policy_dn[256 + 1];
		size_t policy_dn_len = sizeof(policy_dn);

		rc = nmasldap_get_password_policy_dn(ld, object_dn, &policy_dn_len, policy_dn);
		DEBUG(5, ("nmasldap_get_password_policy_dn returned %d\n", rc));
		if (rc == LDAP_SUCCESS) {
			/* We got the password policy dn, check the password against it. */
			DEBUG(5, ("nmasldap_get_password_policy_dn returned %d, %s\n", rc, policy_dn));
			/* We got the password, now check it against the user's password policy */
			rc = nmasldap_policy_check_password(ld, object_dn, pwd);
			DEBUG(3, ("nmasldap_policy_check_password(%s) returned %d\n", object_dn, rc));
			if (rc != LDAP_SUCCESS) {
				/* password failed against current password policy */
				return LDAP_INSUFFICIENT_ACCESS;
			}
		}

		/* No policy dn available, but we got the password return success */
		return LDAP_SUCCESS;
	}

	/* We couldn't get the password */
	return LDAP_INVALID_CREDENTIALS;
}

/**********************************************************************
 Set the users NDS, Universal and Simple passwords.
 ********************************************************************/
int pdb_nds_set_password(struct smbldap_state *ldap_state, char *object_dn, const char *pwd)
{
	LDAP *ld = ldap_state->ldap_struct;
	int rc = -1;
	LDAPMod **tmpmods = NULL;

	if (nmasldap_set_password != NULL) {
		rc = nmasldap_set_password(ld, object_dn, pwd);
		if (rc == LDAP_SUCCESS) {
			DEBUG(5,("NDS Universal Password changed for user %s\n", object_dn));
		} else {
			/* This will fail if Universal Password is not enabled for the user's context */
			DEBUG(3,("NDS Universal Password could not be changed for user %s: %d\n",
					 object_dn, rc));
		}
	}

	if (nmasldap_put_simple_pwd != NULL) {
		rc = nmasldap_put_simple_pwd(ld, object_dn, pwd);
		if (rc == LDAP_SUCCESS) {
			DEBUG(5,("NDS Simple Password changed for user %s\n", object_dn));
		} else {
			/* This will fail if the user's SimplePassword is not set. */
			DEBUG(3,("NDS Simple Password could not be changed for user %s: %d\n",
					 object_dn, rc));
		}
	}

	/* Set eDirectory Password */
	smbldap_set_mod(&tmpmods, LDAP_MOD_REPLACE, "userPassword", pwd);
	rc = smbldap_modify(ldap_state, object_dn, tmpmods);

	return rc;
}

/**********************************************************************
 Load NMAS LDAP Extension library.
 *********************************************************************/
void pdb_nds_unload_nmas_ext_lib(void)
{
	/* Unload NMAS LDAP Extension library */
	if(nmasldap_ext_lib_handle != NULL) {
		sys_dlclose(nmasldap_ext_lib_handle);
		nmasldap_ext_lib_handle = NULL;
		DEBUG(3, ("free_private_data() %s unloaded.\n", NMASLDAP_EXT_LIB));
	}
}

/**********************************************************************
 Load NMAS LDAP Extension library and import functions.
 *********************************************************************/
BOOL pdb_nds_load_nmas_ext_lib(void)
{
	char *err = NULL;

	/* Load NMAS LDAP Extension library. */
	nmasldap_ext_lib_handle = sys_dlopen(NMASLDAP_EXT_LIB, RTLD_NOW | RTLD_GLOBAL);
	if(nmasldap_ext_lib_handle == NULL) {
		DEBUG(0, ("pdb_nds_load_nmas_ext_lib: failed to load '%s'\n", NMASLDAP_EXT_LIB));
		return False;
	}

	/* Import password get function: nmasldap_get_password */
	dlerror(); /* Clear any existing error */
	nmasldap_get_password = sys_dlsym(nmasldap_ext_lib_handle, NMASLDAP_GET_PASSWORD);
	if ((err = dlerror()) != NULL) {
		DEBUG(0, ("pdb_nds_load_nmas_ext_lib: failed to import symbol '%s'\n", NMASLDAP_GET_PASSWORD));
	}

	/* Import password set function: nmasldap_set_password */
	dlerror(); /* Clear any existing error */
	nmasldap_set_password = sys_dlsym(nmasldap_ext_lib_handle, NMASLDAP_SET_PASSWORD);
	if ((err = dlerror()) != NULL) {
		DEBUG(0, ("pdb_nds_load_nmas_ext_lib: failed to import symbol '%s'\n", NMASLDAP_SET_PASSWORD));
	}

	/* Import simple password get function: nmasldap_get_simple_pwd */
	dlerror(); /* Clear any existing error */
	nmasldap_get_simple_pwd = sys_dlsym(nmasldap_ext_lib_handle, NMASLDAP_GET_SIMPLE_PWD);
	if ((err = dlerror()) != NULL) {
		DEBUG(0, ("pdb_nds_load_nmas_ext_lib: failed to import symbol '%s'\n", NMASLDAP_GET_SIMPLE_PWD));
	}

	/* Import simple password set function: nmasldap_put_simple_pwd */
	dlerror(); /* Clear any existing error */
	nmasldap_put_simple_pwd = sys_dlsym(nmasldap_ext_lib_handle, NMASLDAP_PUT_SIMPLE_PWD);
	if ((err = dlerror()) != NULL) {
		DEBUG(0, ("pdb_nds_load_nmas_ext_lib: failed to import symbol '%s'\n", NMASLDAP_PUT_SIMPLE_PWD));
	}

	/* Import password check function: nmasldap_get_password_policy_dn */
	dlerror(); /* Clear any existing error */
	nmasldap_get_password_policy_dn = sys_dlsym(nmasldap_ext_lib_handle, NMASLDAP_GET_PASSWORD_POLICY_DN);
	if ((err = dlerror()) != NULL) {
		DEBUG(0, ("pdb_nds_load_nmas_ext_lib: failed to import symbol '%s'\n", NMASLDAP_GET_PASSWORD_POLICY_DN));
	}

	/* Import password check function: nmasldap_policy_check_password */
	dlerror(); /* Clear any existing error */
	nmasldap_policy_check_password = sys_dlsym(nmasldap_ext_lib_handle, NMASLDAP_POLICY_CHECK_PASSWORD);
	if ((err = dlerror()) != NULL) {
		DEBUG(0, ("pdb_nds_load_nmas_ext_lib: failed to import symbol '%s'\n", NMASLDAP_POLICY_CHECK_PASSWORD));
	}

	/* We must successfully import at least one method for GETTING the password */
	if ((nmasldap_get_password) && (nmasldap_get_simple_pwd))
		return False;

	/* We must successfully import at least one method for SETTING the password */
	if ((nmasldap_set_password) && (nmasldap_put_simple_pwd))
		return False;

	return True;
}

